// Reset = PortB.6
// CS    = PortB.5   (~SS)
// GND   = GND
// SData = PortB.2   (MOSI)
// SClk  = PORTB.1   (SCK)

#include <inttypes.h>
#include <stdio.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include "disp3k.h"

#define SW_SPI
/*
 * global variables
 */
uint8_t color=0;
uint8_t xCoord=0;
uint8_t yCoord=0;
uint8_t memAccessCtrl=0;
uint8_t fontOffset=0;

font f;

void delay5ms(void) {
	volatile uint16_t i;
	for(i=0; i<2200; i++);
}

void delay(void) {
	volatile uint16_t i;
	for(i=0; i<4; i++);
}

#if defined SW_SPI
void ShiftData(uint8_t data)
{
	uint8_t i=8;
  delay(); // small delay
	while (i > 0)
	{		
		PORTB &= ~(0x01 << SCLK);
		if (data >> (i-1))
			(PORTB |= 0x01 << SDATA);
		else
			(PORTB &= ~(0x01 << SDATA));
		data &= ~(1 << (i-1));		
		i--;
		PORTB |= 0x01 << SCLK;
		delay(); // small delay;
	}		
}
#endif

void disp3kSendCommand(uint8_t cmd) {
	

	// CS auf Low setzen
	PORTB &= ~(0x01 << CS);
	
	// 1. Bit (Command/Data) per Software erzeugen
	delay(); // small delay
	PORTB &= ~(0x01 << SCLK);
	delay(); // small delay
	PORTB &= ~(0x01 << SDATA);		// logisch 0 fr Command
	delay(); // small delay
	PORTB |= 0x01 << SCLK;

#if defined SW_SPI
	ShiftData(cmd);
#else

	// Command byte bers SPI senden
	SPCR = (0x01 << SPE) | (0x01 << MSTR);// | (0x01 << SPR1) | (0x01 << SPR0); // 
	SPDR = cmd;
	
	// Auf bertragung warten und CS wieder auf High setzen
	while(!(SPSR & (0x01 << SPIF)));
	SPCR = 0x00;					// SPI wieder disablen

#endif	
	PORTB &= ~(0x01 << SCLK);  // clock low
	delay(); // small delay
	PORTB |= 0x01 << CS;			// CS wieder auf 1
	
}

void disp3kSendData(uint8_t data) {
#ifdef DEBUG
	volatile uint8_t i;
#endif
  

	// CS auf Low setzen
	PORTB &= ~(0x01 << CS);
	delay(); // small delay
	// 1. Bit (Command/Data) per Software erzeugen
	PORTB &= ~(0x01 << SCLK);
  delay(); // small delay
	PORTB |= 0x01 << SDATA;		// logisch 1 fr Daten
  delay(); // small delay
	PORTB |= 0x01 << SCLK;

#if defined SW_SPI
	ShiftData(data);
#else

	// Command byte bers SPI senden
	SPCR = (0x01 << SPE) | (0x01 << MSTR);// | (0x01 << SPR1) | (0x01 << SPR0); // 
	SPDR = data;
	
	// Auf bertragung warten und CS wieder auf High setzen
	while(!(SPSR & (0x01 << SPIF)));
	SPCR = 0x00;					// SPI wieder disablen			
	

#endif	

	PORTB &= ~(0x01 << SCLK);  // clock low
	delay(); // small delay
	PORTB |= 0x01 << CS;			// CS wieder auf 1

}

void disp3kInit(void) {
	uint8_t i;
	
	DDRB |= 0x66;  // set CLK, DATA, CS and RS as output 
  	
	//SPSR |= 0x01 << SPI2X; // Double SPI Speed
	
	// LCD Hardware Reset
	PORTB |= 0x01 << CS;
	PORTB |= 0x01 << RESET;	
	PORTB |= 0x01 << SDATA;
	PORTB |= 0x01 << SCLK;
	PORTB |= 0x01 << CS;
	delay5ms();
	PORTB &= ~(0x01 << SCLK);  // clock low at start
	delay5ms();
	disp3kSendCommand(SOFT_RESET);
	disp3kSendCommand(SLEEP_OUT);
	disp3kSendCommand(DISPLAY_ON);
	disp3kSendCommand(BOOSTER_ON);
	// Booster Voltage stabilisieren lassen
	for(i=0; i<5; i++) {
		delay5ms();
	}
	disp3kSendCommand(COLOR_INTERFACE);
	disp3kSendData(COLOR_8_BIT);
	// initialize Look Up Table for 8 bit mode
	disp3kSendCommand(COLOR_SET);
	// red
	disp3kSendData(0x00);
	disp3kSendData(0x02);
	disp3kSendData(0x03);
	disp3kSendData(0x04);
	disp3kSendData(0x05);
	disp3kSendData(0x06);
	disp3kSendData(0x08);
	disp3kSendData(0x0F);
	// green
	disp3kSendData(0x00);
	disp3kSendData(0x02);
	disp3kSendData(0x03);
	disp3kSendData(0x04);
	disp3kSendData(0x05);
	disp3kSendData(0x06);
	disp3kSendData(0x08);
	disp3kSendData(0x0F);
	// blue
	disp3kSendData(0x00);
	disp3kSendData(0x03);
	disp3kSendData(0x06);
	disp3kSendData(0x0F);
	disp3kSendCommand(MEM_ACCESS_CTRL);
	memAccessCtrl = (0x01 << MIRROR_X) | (0x01 << RGB);
	disp3kSendData(memAccessCtrl);
	disp3kClear(); 
}

void disp3kClear(void) {
	uint16_t i;
	
	disp3kSetDrawRect(DISP_X_START, DISP_X_END, DISP_Y_START, DISP_Y_END);
	disp3kSendCommand(MEM_WRITE);
	for(i=0; i<17030; i++) {
		disp3kSendData(WHITE);
	}
}

int disp3kDrawChar(char c) {
	uint16_t index;
	uint8_t i, j, w, data, skipped=1, startY=yCoord;
	
	disp3kSendCommand(MEM_ACCESS_CTRL);
	memAccessCtrl |= (0x01 << VERT_WRITE);
	disp3kSendData(memAccessCtrl);
	
	if(c == '\n')
		disp3kNewLine();
	if(c < 32)
		return 0;
	
	c -= 32;
	index = c*f.width*(f.height/8);
	
	for(w=0; w<f.width; w++) {
		for(i=0; i<f.height/8; i++) {
			data = pgm_read_byte(f.charData + index++);
			for(j=0; j<8; j++) {
				if(data & 0x01) {
					if(skipped) {
						disp3kGotoXY(xCoord, yCoord);
						disp3kSendCommand(MEM_WRITE);
						skipped=0;
					}
					disp3kDrawPixel();
					yCoord++;
				} else {
					skipped=1;
					disp3kGotoXY(xCoord, ++yCoord);
				}
				data >>= 1;
			}
		}
		disp3kGotoXY(++xCoord, startY);
	}
	
	disp3kSendCommand(MEM_ACCESS_CTRL);
	memAccessCtrl &= ~(0x01 << VERT_WRITE);
	disp3kSendData(memAccessCtrl);
	
	return 0;
}

void disp3kNewLine(void) {
	if(yCoord+f.height < 128)
		disp3kGotoXY(fontOffset, yCoord+f.height);
	else
		disp3kGotoXY(fontOffset, 0);
}

void disp3kDrawCircle(uint8_t xCenter, uint8_t yCenter, uint8_t radius) {
  	int16_t tSwitch, y, x = 0;
  	uint8_t d;

  	d = yCenter - xCenter;
  	y = radius;
  	tSwitch = 3 - 2 * radius;
  	
  	while (x <= y) {
    	disp3kDrawPixelXY(xCenter + x, yCenter + y);
    	disp3kDrawPixelXY(xCenter + x, yCenter - y);
    	
    	disp3kDrawPixelXY(xCenter - x, yCenter + y);
    	disp3kDrawPixelXY(xCenter - x, yCenter - y);
    	
    	disp3kDrawPixelXY(yCenter + y - d, yCenter + x);
    	disp3kDrawPixelXY(yCenter + y - d, yCenter - x);
    	
    	disp3kDrawPixelXY(yCenter - y - d, yCenter + x);
    	disp3kDrawPixelXY(yCenter - y - d, yCenter - x);

    	if (tSwitch < 0) {
    		tSwitch += (4 * x + 6);
    	} else {
      		tSwitch += (4 * (x - y) + 10);
      		y--;
    	}
    	x++;
  	}
}

void disp3kDrawLine(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2) {
	uint8_t length, xTmp, yTmp, i, y, yAlt;
	int16_t m;
	
	if(x1 == x2) {									// vertical line
		// x1|y1 must be the upper point
		if(y1 > y2) {
			xTmp = x1;
			yTmp = y1;
			x1 = x2;
			y1 = y2;
			x2 = xTmp;
			y2 = yTmp;
		}
		
		length = y2-y1;
		for(i=0; i<=length; i++) {
			disp3kDrawPixelXY(x1, y1+i);
		}
	} else if(y1 == y2) {							// horizontal line
		// x1|y1 must be the left point
		if(x1 > x2) {
			xTmp = x1;
			yTmp = y1;
			x1 = x2;
			y1 = y2;
			x2 = xTmp;
			y2 = yTmp;
		}
		
		length = x2-x1;
		for(i=0; i<=length; i++) {
			disp3kDrawPixelXY(x1+i, y1);
		}
	} else {
		// x1 must be smaller than x2
		if(x1 > x2) {
			xTmp = x1;
			yTmp = y1;
			x1 = x2;
			y1 = y2;
			x2 = xTmp;
			y2 = yTmp;
		}
		
		if((y2-y1) >= (x2-x1) || (y1-y2) >= (x2-x1)) {	// angle larger or equal 45
			length = x2-x1;								// not really the length :)
			m = ((y2-y1)*200)/length;
			yAlt = y1;
			for(i=0; i<=length; i++) {
				y = ((m*i)/200)+y1;
				
				if((m*i)%200 >= 100)
					y++;
				else if((m*i)%200 <= -100)
					y--;
				
				disp3kDrawLine(x1+i, yAlt, x1+i, y);
				if(length<=(y2-y1) && y1 < y2)
					yAlt = y+1;
				else if(length<=(y1-y2) && y1 > y2)
					yAlt = y-1;
				else
					yAlt = y;
			}
		} else {										// angle smaller 45
			// y1 must be smaller than y2
			if(y1 > y2) {
				xTmp = x1;
				yTmp = y1;
				x1 = x2;
				y1 = y2;
				x2 = xTmp;
				y2 = yTmp;
			}
			length = y2-y1;
			m = ((x2-x1)*200)/length;
			yAlt = x1;
			for(i=0; i<=length; i++) {
				y = ((m*i)/200)+x1;
				
				if((m*i)%200 >= 100)
					y++;
				else if((m*i)%200 <= -100)
					y--;
				
				disp3kDrawLine(yAlt, y1+i, y, y1+i);
				if(length<=(x2-x1) && x1 < x2)
					yAlt = y+1;
				else if(length<=(x1-x2) && x1 > x2)
					yAlt = y-1;
				else
					yAlt = y;
			}
		}
	}
}

void disp3kDrawPixel(void) {
	disp3kSendData(color);
}

void disp3kDrawPixelXY(uint8_t x, uint8_t y) {
	disp3kGotoXY(x, y);
	disp3kSendCommand(MEM_WRITE);
	disp3kSendData(color);
}

void disp3kDrawRect(uint8_t x, uint8_t y, uint8_t width, uint8_t height) {
	width--;
	height--;
	disp3kDrawLine(x, y, x+width, y);				// top
	disp3kDrawLine(x, y, x, y+height);				// left
	disp3kDrawLine(x, y+height, x+width, y+height);	// bottom
	disp3kDrawLine(x+width, y, x+width, y+height);	// right
}

void disp3kFillRect(uint8_t x, uint8_t y, uint8_t width, uint8_t height) {
	uint16_t i;
	
	disp3kSetDrawRect(x, x+width-1, y, y+height-1);
	disp3kSendCommand(MEM_WRITE);
	for(i=0; i<width*height; i++) {
		disp3kSendData(color);
	}
}

void disp3kGotoXY(uint8_t x, uint8_t y) {
	disp3kSetDrawRect(x, DISP_X_END, y, DISP_Y_END);
}

void disp3kSetColor(uint8_t newColor) {
	color = newColor;
}

void disp3kSetFont(font newFont) {
	f = newFont;
}

void disp3kSetFontOffset(uint8_t newOffset) {
	fontOffset = newOffset;
}

void disp3kSetDrawRect(uint8_t startX, uint8_t endX, uint8_t startY, uint8_t endY) {
	disp3kSendCommand(SET_X_ADDR);
	disp3kSendData(startX+1);
	disp3kSendData(endX+1);
	
	disp3kSendCommand(SET_Y_ADDR);
	disp3kSendData(startY+1);
	disp3kSendData(endY+1);
	
	xCoord=startX;
	yCoord=startY;
}

/* display text string */
void disp3kprintf(uint8_t *string)
{
	 uint16_t i;
	 while(string[i++] != 0)
	   disp3kDrawChar(string[i]);	  
}

/*
 * Draws a nice little test screen
 */
void disp3kTestScreen(void) {
	uint16_t i;
	uint8_t f, j, r, g, b;
	
	disp3kSetDrawRect(DISP_X_START, DISP_X_END, DISP_Y_START, DISP_Y_END);
	disp3kSendCommand(MEM_WRITE);
	
	// schwarz -> rot fade
	for(j=0; j<8; j++) {
		f=0;
		for(i=0; i<130; i++) {
			r = j;
			g = 0;
			b = 0;
			
			if(i > 0 && i % 15 == 0) {
				f++;
			}
			if(r+f < 8) r+=f; else r=7;
			if(g+f < 8) g+=f; else g=7;
			if(b+(f/2) < 4) b+=(f/2); else b=3;
			
			disp3kSendData((r << 5) | (g << 2) | b);
		}
	}
	
	// rot -> gelb fade
	for(j=0; j<8; j++) {
		f=0;
		for(i=0; i<130; i++) {
			r = 7;
			g = j;
			b = 0;
			
			if(i > 0 && i % 15 == 0) {
				f++;
			}
			if(r+f < 8) r+=f; else r=7;
			if(g+f < 8) g+=f; else g=7;
			if(b+(f/2) < 4) b+=(f/2); else b=3;
			
			disp3kSendData((r << 5) | (g << 2) | b);
		}
	}
	
	// gelb -> schwarz fade
	for(j=0; j<8; j++) {
		f=0;
		for(i=0; i<130; i++) {
			r = 7-j;
			g = 7-j;
			b = 0;
			
			if(i > 0 && i % 15 == 0) {
				f++;
			}
			if(r+f < 8) r+=f; else r=7;
			if(g+f < 8) g+=f; else g=7;
			if(b+(f/2) < 4) b+=(f/2); else b=3;
			
			disp3kSendData((r << 5) | (g << 2) | b);
		}
	}
	
	// schwarz -> grn fade
	for(j=0; j<8; j++) {
		f=0;
		for(i=0; i<130; i++) {
			r = 0;
			g = j;
			b = 0;
			
			if(i > 0 && i % 15 == 0) {
				f++;
			}
			if(r+f < 8) r+=f; else r=7;
			if(g+f < 8) g+=f; else g=7;
			if(b+(f/2) < 4) b+=(f/2); else b=3;
			
			disp3kSendData((r << 5) | (g << 2) | b);
		}
	}
	
	// grn -> cyan fade
	for(j=0; j<8; j++) {
		f=0;
		for(i=0; i<130; i++) {
			r = 0;
			g = 7;
			b = j;
			
			if(i > 0 && i % 15 == 0) {
				f++;
			}
			if(r+f < 8) r+=f; else r=7;
			if(g+f < 8) g+=f; else g=7;
			if(b+(f/2) < 4) b+=(f/2); else b=3;
			
			disp3kSendData((r << 5) | (g << 2) | b);
		}
	}
	
	// cyan -> schwarz fade
	for(j=0; j<8; j++) {
		f=0;
		for(i=0; i<130; i++) {
			r = 0;
			g = 7-j;
			b = 7-j;
			
			if(i > 0 && i % 15 == 0) {
				f++;
			}
			if(r+f < 8) r+=f; else r=7;
			if(g+f < 8) g+=f; else g=7;
			if(b+(f/2) < 4) b+=(f/2); else b=3;
			
			disp3kSendData((r << 5) | (g << 2) | b);
		}
	}
	
	// schwarz -> blau fade
	for(j=0; j<8; j++) {
		f=0;
		for(i=0; i<130; i++) {
			r = 0;
			g = 0;
			b = j/2;
			
			if(i > 0 && i % 15 == 0) {
				f++;
			}
			if(r+f < 8) r+=f; else r=7;
			if(g+f < 8) g+=f; else g=7;
			if(b+(f/2) < 4) b+=(f/2); else b=3;
			
			disp3kSendData((r << 5) | (g << 2) | b);
		}
	}
	
	// blau -> magenta fade
	for(j=0; j<8; j++) {
		f=0;
		for(i=0; i<130; i++) {
			r = j;
			g = 0;
			b = 3;
			
			if(i > 0 && i % 15 == 0) {
				f++;
			}
			if(r+f < 8) r+=f; else r=7;
			if(g+f < 8) g+=f; else g=7;
			if(b+(f/2) < 4) b+=(f/2); else b=3;
			
			disp3kSendData((r << 5) | (g << 2) | b);
		}
	}
	
	// magenta -> schwarz fade
	for(j=0; j<8; j++) {
		f=0;
		for(i=0; i<130; i++) {
			r = 7-j;
			g = 0;
			b = 3-j/2;
			
			if(i > 0 && i % 15 == 0) {
				f++;
			}
			if(r+f < 8) r+=f; else r=7;
			if(g+f < 8) g+=f; else g=7;
			if(b+(f/2) < 4) b+=(f/2); else b=3;
			
			disp3kSendData((r << 5) | (g << 2) | b);
		}
	}
	
	for(i=0; i<260; i++) {
		disp3kSendData(0xFF);		// weier strich
	}
	for(i=0; i<1040; i++) {
		disp3kSendData(0xE0);		// rot
	}
	for(i=0; i<1040; i++) {
		disp3kSendData(0x1C);		// grn
	}
	for(i=0; i<1040; i++) {
		disp3kSendData(0x03);		// blau
	}
	for(i=0; i<1040; i++) {
		disp3kSendData(0xFC);		// gelb
	}
	for(i=0; i<1040; i++) {
		disp3kSendData(0xE3);		// magenta
	}
	for(i=0; i<1040; i++) {
		disp3kSendData(0x1F);		// cyan
	}
	for(i=0; i<1040; i++) {
		disp3kSendData(0x00);		// schwarz
	}
}
