
#ifndef DISP3K_H
#define DISP3K_H

/*
 * Pins
 * CS		-> frei definierbar, sollte auf ~SS vom SPI liegen
 * SDATA	-> MOSI (SPI)
 * SCLK		-> CLK (SPI)
 * RESET	-> frei definierbar
 */
#define CS		5
#define SCLK	1
#define SDATA	2
#define RESET	6

//#define DEBUG

/*
 * Commands
 */
#define SOFT_RESET		0x01
#define SLEEP_OUT		0x11
#define DISPLAY_ON		0x29
#define BOOSTER_ON		0x03
#define COLOR_INTERFACE	0x3A
#define COLOR_8_BIT	0x02
#define COLOR_SET		0x2D
#define MEM_ACCESS_CTRL	0x36
#define MIRROR_X		6
#define MIRROR_Y		7
#define VERT_WRITE	5
#define RGB			3
#define MEM_WRITE		0x2C
#define SET_X_ADDR		0x2A
#define SET_Y_ADDR		0x2B

/*
 * for disp3kSetDrawRect();
 */
#define DISP_X_START	0
#define DISP_Y_START	0
#define DISP_X_END	129
#define DISP_Y_END	129

/*
 * a few basic colors
 */
#define RED				0xE0
#define GREEN			0x1C
#define BLUE			0x03
#define YELLOW			0xFC
#define MAGENTA			0xE3
#define CYAN			0x1F
#define BLACK			0x00
#define WHITE			0xFF

#define printf(x) disp3kprintf(x);

struct font {
	uint8_t width;
	uint8_t height;
	PGM_P charData;
};
typedef struct font font;

typedef enum
{
	CMD = 0,
	DATA
}DataType;

/*
 * Function Prototypes
 */
int disp3kDrawChar(char c);
void disp3kClear(void);
void disp3kDrawCircle(uint8_t xCenter, uint8_t yCenter, uint8_t radius);
void disp3kDrawLine(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2);
void disp3kDrawPixel(void);
void disp3kDrawPixelXY(uint8_t x, uint8_t y);
void disp3kDrawRect(uint8_t x, uint8_t y, uint8_t width, uint8_t height);
void disp3kFillRect(uint8_t x, uint8_t y, uint8_t width, uint8_t height);
void disp3kGotoXY(uint8_t x, uint8_t y);
void disp3kInit(void);
void disp3kNewLine(void);
void disp3kSendCommand(uint8_t command);
void disp3kSendData(uint8_t data);
void disp3kSetColor(uint8_t newColor);
void disp3kSetFont(font newFont);
void disp3kSetFontOffset(uint8_t newOffset);
void disp3kSetDrawRect(uint8_t startX, uint8_t endX, uint8_t startY, uint8_t endY);
void disp3kTestScreen(void);
void disp3kprintf(uint8_t *string);

#endif
